import unittest
from geojson import Point

import frost_sta_client.model.location
import frost_sta_client.model.thing
import frost_sta_client.utils
import frost_sta_client.model.ext.entity_list
from frost_sta_client.model.ext import entity_type


class TestEntityFormatter(unittest.TestCase):
    
    def test_create_thing_basic(self):
        exp_result = dict(name='nice thing',
                          description='the description of the thing',
                          properties=dict(
                                    nice=True,
                                    level_of_niceness=1000
                                )
                          )
        entity = frost_sta_client.model.thing.Thing()
        entity.name = 'nice thing'
        entity.description = 'the description of the thing'
        properties = {'nice': True, 'level_of_niceness': 1000}
        entity.properties = properties
        entity_json = frost_sta_client.utils.transform_entity_to_json_dict(entity)
        self.assertDictEqual(exp_result, entity_json)

    def test_write_thing_completely_empty(self):
        result = {}

        entity = frost_sta_client.model.thing.Thing()
        entity_json = frost_sta_client.utils.transform_entity_to_json_dict(entity)
        self.assertDictEqual(result, entity_json)

    def test_write_thing_with_location(self):
        result = {'name': 'another nice thing',
                  'description': 'This thing has also a nice location',
                  'Locations': [
                      {
                          '@iot.id': 1,
                      }
                  ]}
        entity = frost_sta_client.model.thing.Thing()
        entity.name = 'another nice thing'
        entity.description = 'This thing has also a nice location'
        entity.properties = {}

        location = frost_sta_client.model.location.Location()
        location.id = 1
        entity.locations = frost_sta_client.model.ext.entity_list.EntityList(entities=[location],
                                                                             entity_class=entity_type.EntityTypes['Location']['class'])
        entity_json = frost_sta_client.utils.transform_entity_to_json_dict(entity)
        self.assertDictEqual(result, entity_json)


    def test_write_thing_with_specified_id_and_attributes(self):
        result = {'@iot.id': 123,
                  'name': 'another nice thing',
                  'Locations': [
                      {
                          '@iot.id': 456,
                          'name': 'location with specified id'
                      }
                  ],
                  'HistoricalLocations': [
                      {
                          '@iot.id': 789
                      }
                  ],
                  'Datastreams': [
                      {
                          'name': 'Datastream without specified id'
                      }
                  ],
                  }
        entity = frost_sta_client.model.thing.Thing(id=123)
        entity.name = 'another nice thing'

        location = frost_sta_client.model.location.Location(name='location with specified id')
        location.id = 456
        historical_location = frost_sta_client.model.historical_location.HistoricalLocation(id=789)
        datastream = frost_sta_client.model.datastream.Datastream(name='Datastream without specified id')

        entity.locations = frost_sta_client.model.ext.entity_list.EntityList(entities=[location],
                                                                             entity_class=entity_type.EntityTypes['Location']['class'])
        entity.historical_locations = frost_sta_client.model.ext.entity_list.EntityList(entities=[historical_location],
                                                                             entity_class=entity_type.EntityTypes['HistoricalLocation']['class'])
        entity.datastreams = frost_sta_client.model.ext.entity_list.EntityList(entities=[datastream], entity_class=entity_type.EntityTypes['Datastream']['class'])
        entity_json = frost_sta_client.utils.transform_entity_to_json_dict(entity)
        self.assertDictEqual(result, entity_json)


    def test_incorrect_collection(self):
        exp_result = {
            'name': 'test thing',
            'description': 'incorrect thing for testing',
            'Locations': [
                {
                    'name': 'favorite place',
                    'description': 'this is my favorite place',
                    'encodingType': 'application/vnd.geo+json',
                    'location': {
                        'type': 'Point',
                        'coordinates': [-49.593, 85.23]
                    }
                }
            ]
        }

        entity = frost_sta_client.model.thing.Thing()
        entity.name = 'test thing'
        entity.description = 'incorrect thing for testing'
        
        location = frost_sta_client.model.location.Location()
        location.name = 'favorite place'
        location.description = 'this is my favorite place'
        location.encoding_type = 'application/vnd.geo+json'
        location.location = Point((-49.593, 85.23))
        
        entity.locations = frost_sta_client.model.ext.entity_list.EntityList(entities=[location],
                                                                             entity_class=entity_type.EntityTypes['Location']['class'])
        entity_json = frost_sta_client.utils.transform_entity_to_json_dict(entity)
        self.assertDictEqual(exp_result, entity_json)

    def test_write_location_geojson(self):
        input_json = {
            '@iot.id': 1,
            'name': 'Treasure',
            'description': 'location of the treasure',
            'encodingType': 'application/vnd.geo+json',
            'location': {
                'type': 'Point',
                'coordinates': [-49.593, 85.23]
            }
        }
        exp_result = frost_sta_client.model.location.Location()
        exp_result.id = 1
        exp_result.name = 'Treasure'
        exp_result.description = 'location of the treasure'
        exp_result.encoding_type = 'application/vnd.geo+json'
        exp_result.location = Point((-49.593, 85.23))
        result = frost_sta_client.utils.transform_json_to_entity(input_json, 'frost_sta_client.model.location.Location')
        self.assertEqual(result, exp_result)


if __name__ == '__main__':
    unittest.main()



__title__ = 'frost_sta_client'
__version__ = '1.1.38'
__license__ = 'LGPL3'
__author__ = 'Jonathan Vogl'
__copyright__ = 'Fraunhofer IOSB'
__contact__ = 'jonathan.vogl@iosb.fraunhofer.de'
__url__ = 'https://github.com/FraunhoferIOSB/FROST-Python-Client'
__description__ = 'a client library to facilitate interaction with a FROST SensorThingsAPI Server'



import unittest

from frost_sta_client.utils import parse_datetime
import datetime


class TestUtils(unittest.TestCase):

    def test_parse_iso_time_with_timezone(self):
        parsedtime = parse_datetime('2022-04-07T14:00:00+02:00')
        self.assertEqual('2022-04-07T14:00:00+02:00', parsedtime)

    def test_parse_iso_time(self):
        parsedtime = parse_datetime('2022-04-07T14:00:00Z')
        self.assertEqual('2022-04-07T14:00:00+00:00', parsedtime)

    def test_parse_interval(self):
        parsedtime = parse_datetime('2022-04-07T14:00:00Z/2022-04-07T15:00:00Z')
        self.assertEqual('2022-04-07T14:00:00+00:00/2022-04-07T15:00:00+00:00', parsedtime)

    def test_parse_interval_with_timezone_offset(self):
        parsedtime = parse_datetime('2022-04-07T14:00:00+02:00/2022-04-07T15:00:00+02:00')
        self.assertEqual('2022-04-07T14:00:00+02:00/2022-04-07T15:00:00+02:00', parsedtime)




# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from frost_sta_client.dao import base
from frost_sta_client.model.ext.entity_type import EntityTypes
from frost_sta_client.utils import transform_entity_to_json_dict
import frost_sta_client

import logging
import requests
import json



class ObservationDao(base.BaseDao):
    CREATE_OBSERVATIONS = "CreateObservations"

    def __init__(self, service):
        """
        A data access object for operations with the Observation entity
        """
        base.BaseDao.__init__(self, service, EntityTypes["Observation"])

    def create(self, entity):
        if isinstance(entity, frost_sta_client.model.observation.Observation):
            super().create(entity)
        else:
            # entity is probably a data array
            url = self.service.url.copy()
            url.path.add(self.CREATE_OBSERVATIONS)
            logging.debug('Posting to ' + str(url.url))
            json_dict = transform_entity_to_json_dict(entity.value)
            try:
                response = self.service.execute('post', url, json=json_dict)
            except requests.exceptions.HTTPError as e:
                error_json = e.response.json()
                error_message = error_json['message']
                logging.error("Creating {} failed with status-code {}, {}".format("Data Array",
                                                                              e.response.status_code,
                                                                              error_message))
            response_text_as_list = json.loads(response.text)
            result = [frost_sta_client.model.observation.Observation(self_link=link) for link in response_text_as_list]
            return result



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import requests
import frost_sta_client


class EntityList:
    def __init__(self, entity_class, entities=None):
        if entities is None:
            entities = []
        self.entities = entities
        self.entity_class = entity_class
        self.next_link = None
        self.service = None
        self.iterable_entities = None
        self.count = None
        self.callback = None
        self.step_size = None

    def __new__(cls, *args, **kwargs):
        new_entity_list = super().__new__(cls)
        attributes = {'_entities': None, '_entity_class': '', '_next_link': '', '_service': {}, '_count': '',
                      '_iterable_entities': None, '_callback': None,
                      '_step_size': None}
        for key, value in attributes.items():
            new_entity_list.__dict__[key] = value
        return new_entity_list

    def __iter__(self):
        self.iterable_entities = iter(enumerate(self.entities))
        return self

    def __next__(self):
        idx, next_entity = next(self.iterable_entities, (len(self.entities), None))
        if self.step_size is not None and idx is not None and idx % self.step_size == 0:
            self.callback(idx)
        if next_entity is not None:
            return next_entity
        if self.next_link is not None:
            try:
                response = self.service.execute('get', self.next_link)
            except requests.exceptions.HTTPError as e:
                error_json = e.response.json()
                error_message = error_json['message']
                logging.error("Query failed with status-code {}, {}".format(e.response.status_code, error_message))
                raise e
            logging.debug('Received response: {} from {}'.format(response.status_code, self.next_link))
            try:
                json_response = response.json()
            except ValueError:
                raise ValueError('Cannot find json in http response')

            result_list = frost_sta_client.utils.transform_json_to_entity_list(json_response, self.entity_class)
            self.entities += result_list.entities
            self.set_service(self.service)
            self.next_link = json_response.get("@iot.nextLink", None)
            self.iterable_entities = iter(enumerate(self.entities[-len(result_list.entities):], start=idx))
            return next(self.iterable_entities)[1]
        raise StopIteration

    def get(self, index):
        if not isinstance(index, int):
            raise IndexError('index must be an integer')
        if index >= len(self.entities):
            raise IndexError('index exceeds total number of entities')
        if index < 0:
            raise IndexError('negative indices cannot be accessed')
        return self.entities[index]

    @property
    def entity_class(self):
        return self._entity_class

    @entity_class.setter
    def entity_class(self, value):
        if isinstance(value, str):
            self._entity_class = value
            return
        raise ValueError('entity_class should be of type str')

    @property
    def entities(self):
        return self._entities

    @entities.setter
    def entities(self, values):
        if isinstance(values, list) and all(isinstance(v, frost_sta_client.model.entity.Entity) for v in values):
            self._entities = values
            return
        raise ValueError('entities should be a list of entities')

    @property
    def callback(self):
        return self._callback

    @callback.setter
    def callback(self, callback):
        if callable(callback) or callback is None:
            self._callback = callback

    @property
    def step_size(self):
        return self._step_size

    @step_size.setter
    def step_size(self, value):
        if isinstance(value, int) or value is None:
            self._step_size = value
            return
        raise ValueError('step_size should be of type int')

    @property
    def next_link(self):
        return self._next_link

    @next_link.setter
    def next_link(self, value):
        if value is None or isinstance(value, str):
            self._next_link = value
            return
        raise ValueError('next_link should be of type string')

    @property
    def service(self):
        return self._service

    @service.setter
    def service(self, value):
        if value is None or isinstance(value, frost_sta_client.service.sensorthingsservice.SensorThingsService):
            self._service = value
            return
        raise ValueError('service should be of type SensorThingsService')

    def set_service(self, service):
        self.service = service
        for entity in self.entities:
            entity.set_service(service)

    def __getstate__(self):
        data = []
        for entity in self.entities:
            data.append(entity.__getstate__())
        return data

    def __setstate__(self, state):
        self._next_link = state.get(self.entities + '@nextLink')
        pass



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import frost_sta_client.model
from frost_sta_client.dao.location import LocationDao

from . import entity
from . import thing
from . import historical_location

import inspect
import json
import geojson

from frost_sta_client import utils
from .ext import entity_type
from .ext import entity_list


class Location(entity.Entity):

    def __init__(self,
                 name='',
                 description='',
                 encoding_type='',
                 properties=None,
                 location=None,
                 things=None,
                 historical_locations=None,
                 **kwargs):
        super().__init__(**kwargs)
        if properties is None:
            properties = {}
        self.name = name
        self.description = description
        self.encoding_type = encoding_type
        self.properties = properties
        self.location = location
        self.things = things
        self.historical_locations = historical_locations

    def __new__(cls, *args, **kwargs):
        new_loc = super().__new__(cls)
        attributes = {'_id': None, '_name': '', '_description': '', '_properties': {}, '_encodingType': '',
                      '_location': None, '_things': None, '_historical_locations': None, '_self_link': '',
                      '_service': None}
        for key, value in attributes.items():
            new_loc.__dict__[key] = value
        return new_loc

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is None:
            self._name = None
            return
        if not isinstance(value, str):
            raise ValueError('name should be of type str!')
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        if value is None:
            self._description = None
            return
        if not isinstance(value, str):
            raise ValueError('description should be of type str!')
        self._description = value

    @property
    def encoding_type(self):
        return self._encoding_type

    @encoding_type.setter
    def encoding_type(self, value):
        if value is None:
            self._encoding_type = None
            return
        if not isinstance(value, str):
            raise ValueError('encodingType should be of type str!')
        self._encoding_type = value

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, values):
        if values is None:
            self._properties = None
            return
        if not isinstance(values, dict):
            raise ValueError('properties should be of type dict!')
        self._properties = values

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        if value is None:
            self._location = None
            return
        geo_classes = [obj for _, obj in inspect.getmembers(geojson) if inspect.isclass(obj) and
                       obj.__module__ == 'geojson.geometry']
        if type(value) in geo_classes:
            self._location = value
            return
        else:
            try:
                json.dumps(value)
            except TypeError:
                raise ValueError('location should be json serializable!')
            self._location = value

    @property
    def things(self):
        return self._things

    @things.setter
    def things(self, values):
        if values is None:
            self._things = None
            return
        if isinstance(values, list) and all(isinstance(th, thing.Thing) for th in values):
            entity_class = entity_type.EntityTypes['Thing']['class']
            self._things = entity_list.EntityList(entity_class=entity_class, entities=values)
            return
        if not isinstance(values, entity_list.EntityList) or \
                any((not isinstance(th, thing.Thing)) for th in values.entities):
            raise ValueError('Things should be a list of things!')
        self._things = values

    @property
    def historical_locations(self):
        return self._historical_locations

    @historical_locations.setter
    def historical_locations(self, values):
        if values is None:
            self._historical_locations = None
            return
        if isinstance(values, list) and all(isinstance(hl, historical_location.HistoricalLocation) for hl in values):
            entity_class = entity_type.EntityTypes['HistoricalLocation']['class']
            self._historical_locations = entity_list.EntityList(entity_class=entity_class, entities=values)
            return
        if isinstance(values, entity_list.EntityList) and \
                all(isinstance(hl, historical_location.HistoricalLocation) for hl in values.entities):
            self._historical_locations = values
        raise ValueError('historical_location should be of type HistoricalLocation!')

    def get_things(self):
        result = self.service.things()
        result.parent = self
        return result

    def get_historical_locations(self):
        result = self.service.historical_locations()
        result.parent = self
        return result

    def ensure_service_on_children(self, service):
        if self.things is not None:
            self.things.set_service(service)
        if self.historical_locations is not None:
            self.historical_locations.set_service(service)

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, type(self)):
            return False
        if id(self) == id(other):
            return True
        if self.name != other.name:
            return False
        if self.description != other.description:
            return False
        if self.encoding_type != other.encoding_type:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        data = super().__getstate__()
        if self.name is not None and self.name != '':
            data['name'] = self.name
        if self.description is not None and self.description != '':
            data['description'] = self.description
        if self.encoding_type is not None and self.encoding_type != '':
            data['encodingType'] = self.encoding_type
        if self.properties is not None and self.properties != {}:
            data['properties'] = self.properties
        if self.location is not None:
            data['location'] = self.location
        if self.things is not None:
            data['Things'] = self.things
        if self.historical_locations is not None and len(self.historical_locations.entities) > 0:
            data['HistoricalLocation'] = self.historical_locations.__getstate__()
        return data

    def __setstate__(self, state):
        super().__setstate__(state)
        self.name = state.get("name", None)
        self.description = state.get("description", None)
        self.encoding_type = state.get("encodingType", None)
        self.properties = state.get("properties", None)
        if state.get("Things", None) is not None:
            entity_class = entity_type.EntityTypes['Thing']['class']
            self.things = utils.transform_json_to_entity_list(state['Things'], entity_class)
            self.things.next_link = state.get('Things@iot.nextLink', None)
            self.things.count = state.get('Things@iot.count', None)
        if state.get("location", None) is not None:
            self.location = state["location"]
        if state.get("HistoricalLocations", None) is not None:
            entity_class = entity_type.EntityTypes['HistoricalLocation']['class']
            self.historical_locations = utils.transform_json_to_entity_list(state['HistoricalLocations'], entity_class)
            self.historical_locations.next_link = state.get('HistoricalLocations@iot.nextLink', None)
            self.historical_locations.count = state.get('HistoricalLocations@iot.count', None)

    def get_dao(self, service):
        return LocationDao(service)



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from frost_sta_client.dao import base
from frost_sta_client.model.ext.entity_type import EntityTypes


class TaskDao(base.BaseDao):
    def __init__(self, service):
        """
        A data access object for operations with the Task entity
        """
        base.BaseDao.__init__(self, service, EntityTypes['Task'])
